/* (non-javadoc)
 * MergeBioAssay.java is part of Merge.
 *
 * Merge is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Merge is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Merge; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * johan.enell@onk.lu.se
 * Johan Enell, Dept Oncology, Lund University, S-221 85 Lund, Sweden
 */

package mergebioassay;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import java.io.FileInputStream;

import mergebioassay.mergers.ArithmeticIntensityMerger;
import mergebioassay.mergers.ArithmeticRatioMerger;
import mergebioassay.mergers.BadPatternException;
import mergebioassay.mergers.GeometricRatioMerger;
import mergebioassay.mergers.Merge_assay;
import mergebioassay.mergers.Merger;
import mergebioassay.mergers.RatioRatioMerger;

import basefile.BASEFileException;
import basefile.BASEFileReader;
import basefile.BASEFileSection;
import basefile.BadSectionException;

/**
 * @author Johan Enell
 * 
 * Merge bioassays is a plugin created for the database BASE and is designed
 * for merging bioassays. There are four merging method implemented in this
 * version which are using both arithmetic and geometric mean to calculte the
 * numbers. They will also allow the choice to calculate the mean from the
 * intensities or form raito.
 */
public class MergeBioAssay
{
	private final int GEOMETRIC_RATIO = 0, ARITHMETIC_RATIO = 1,
			ARITHMETIC_INT = 2, RATIO_RATIO = 3;

	private String param_newname = "";

	private String param_assayname = "%1";

	private String param_group = "%1";

	private String param_annType = "";

	private String param_annValue = "";

	private String param_ratio = "";

	private int param_mergeMethod = GEOMETRIC_RATIO;

	private int nbrOfGroups = -1;

	private int sortnbr = 1;

	private boolean annSort = false;

	private String ratio = "";

	private String ratioNumerator = "";

	private String ratioDenominator = "";

	private NoGroup noGroup = new NoGroup();

	private HashMap<String, List<Merge_assay>> assayGroups = new HashMap<String, List<Merge_assay>>();

	/**
	 * SmoothMain method for the plugin.
	 * 
	 * @param args
	 */
	public static void main(String[] args)
	{
		if (args.length == 1)
		{
			try
			{
				System.setIn(new FileInputStream("stdin.txt"));
				System.setOut(new PrintStream(new FileOutputStream("stdout.txt")));
			}
			catch (FileNotFoundException fnfe)
			{
				fnfe.printStackTrace();
			}
		}

		MergeBioAssay merge = new MergeBioAssay();
		try
		{
			BASEFileReader reader = new BASEFileReader(new File("stdin.txt"));
			BASEFileSection section = null;

			section = reader.readSection(true);
			if (section.isType("merge_bioassays settings"))
			{
				merge.extractSettings(section);
			}

			section = reader.readSection();
			if (section.isType("assays"))
			{
				merge.extractAssays(section, reader);
			}

			section = reader.readSection();
			if (section.isType("spots"))
			{
				merge.mergeSpots(section, reader);
			}

			if (!merge.noGroup.isEmpty())
			{
				System.setIn(new FileInputStream("stdin.txt"));
				reader = new BASEFileReader(new File("stdin.txt"));
				section = reader.readSection();
				while (section != null)
				{
					if (section.isType("spots"))
					{
						merge.noGroup.printNoGroup(section, reader);
						break;
					}
					else
						section = reader.readSection();
				}
			}

			System.setOut(new PrintStream(new FileOutputStream("report.htm")));

			System.out.println("<html><head><title>Report from merge</title></head>\n<body>");
			System.out.println("<big>Parameters</big>");
			System.out.println("<table border=\"0\">");
			System.out.println("<tr>\n<td>assayname\n<td>" + merge.param_assayname);
			System.out.println("<tr>\n<td>group\n<td>" + merge.param_group);
			System.out.println("<tr>\n<td>newname\n<td>" + merge.param_newname);
			System.out.println("<tr>\n<td>ratio\n<td>" + merge.param_ratio);
			System.out.println("<tr>\n<td>mergeMethod\n<td>" + merge.param_mergeMethod);
			System.out.println("<tr>\n<td>annType\n<td>" + merge.param_annType);
			System.out.println("<tr>\n<td>annValue\n<td>" + merge.param_annValue);
			System.out.println("</table>");

			System.out.println("<br>");
			System.out.println("<big>Groups</big>");
			System.out.println("<table border=\"1\">");
			for (String key : merge.assayGroups.keySet())
			{
				System.out.println("<tr>\n<td>" + merge.param_newname + key + "\n<td>");

				List<Merge_assay> group = merge.assayGroups.get(key);
				for (int i = 0; i < group.size(); i++)
				{
					BioAssay tmp = group.get(i);
					System.out.println(tmp.name + "<br>");
				}
			}
			if (!merge.noGroup.isEmpty())
			{
				System.out.println("<tr>\n<td>noGroup<td>");
				for (int i = 0; i < merge.noGroup.size(); i++)
				{
					BioAssay tmp = merge.noGroup.get(i);
					System.out.print(tmp.name + "<br>");
				}
			}
			System.out.println("</table>");

		}
		catch (BASEFileException e)
		{
			e.printStackTrace();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Will read the parameters from the baseFile.
	 * 
	 * @param section - the section from where the parameters will be read
	 * @return true if the parameters was read without any problems
	 */
	private boolean extractSettings(BASEFileSection section)
	{
		param_assayname = section.findStringOpt("assayname");
		param_group = section.findStringOpt("group");
		param_ratio = section.findStringOpt("ratio");
		param_newname = section.findStringOpt("newname");
		param_mergeMethod = section.findIntOpt("mergeMethod");
		param_annType = section.findStringOpt("annType");
		param_annValue = section.findStringOpt("annValue");

		if (param_group.matches("%\\d"))
		{
			if (!param_assayname.matches(".*" + param_group + ".*"))
			{
				System.err.println("MergeBioAssay.extractSettings(BASEFileSection section)");
				System.err.println("Bad argument <sort>: Cant find " + param_group + " in assayname");
				System.exit(0);
			}

			sortnbr = param_group.charAt(1) - 48;
		}
		else
		{
			sortnbr = 10;
			annSort = true;
		}

		if (param_mergeMethod == RATIO_RATIO)
		{
			if (!param_ratio.equals(""))
			{
				String[] tmp = split(param_ratio, param_ratio.charAt(param_ratio.length() - 1));

				if (tmp.length != 3)
				{
					System.err.println("MergeBioAssay.extractSettings(BASEFileSection section)");
					System.err.println("Invalid format of argument <ratio>");
					System.err.println("Must be <ratio>|<numerator>|<denominator>|");
					System.exit(0);
				}
				ratio = tmp[0];
				ratioNumerator = tmp[1];
				ratioDenominator = tmp[2];
			}
			else if (!param_annType.equals("") && !param_annValue.equals(""))
			{
				String[] tmp = split(param_annValue, param_annValue.charAt(param_annValue.length() - 1));
				if (tmp.length != 2)
				{
					System.err.println("MergeBioAssay.extractSettings(BASEFileSection section)");
					System.err.println("Invalid format of argument annValue");
					System.err.println("Can not use delimiter " + param_annValue.charAt(param_annValue.length() - 1));
					System.exit(0);
				}
				ratio = param_annType;
				ratioNumerator = tmp[0];
				ratioDenominator = tmp[1];
			}
			else
			{
				System.err.println("MergeBioAssay.extractSettings(BASEFileSection section)");
				System.err.println("Merge method 'ratio of ratio' chosen but ratio or annotation type is left blank");
				System.exit(0);
			}
		}

		if (param_group.equals(ratio))
		{
			System.err.println("You have the same sort as ratio. This will create crazy values.");
		}
		return true;
	}

	/**
	 * Will read the assay from the baseFile
	 * 
	 * @param section - the section to be read. Must have type "assays"
	 * @param reader - the reader where the section data is located
	 * @return true if the baseFile is read without any problems
	 */
	private boolean extractAssays(BASEFileSection section, BASEFileReader reader)
	{
		int idCol, nameCol, sortCol, nbr = -1, ratioCol = -1;

		List<String> columns = section.findFieldList("columns");
		if (columns == null)
		{
			System.err.println("MergeBioAssay.extractAssays(BASEFileSection section, BASEFileReader reader)");
			System.err.println("Missing header columns in section assays");
			System.exit(0);
		}

		idCol = columns.indexOf("id");
		nameCol = columns.indexOf("name");
		if (idCol == -1 || nameCol == -1)
		{
			System.err.println("MergeBioAssay.extractAssays(BASEFileSection section, BASEFileReader reader)");
			System.err.println("Missing id/name in section assays");
			System.exit(0);
		}

		sortCol = columns.indexOf(param_group);
		if (annSort && sortCol == -1)
		{
			System.err.println("MergeBioAssay.extractAssays(BASEFileSection section, BASEFileReader reader)");
			System.err.println("Missing " + param_group + " in section assays\n");
			System.exit(0);
		}

		if (param_mergeMethod == RATIO_RATIO)
		{

			if (ratio.matches("%\\d"))
			{
				nbr = ratio.charAt(1) - 48;
			}
			else
			{
				ratioCol = columns.indexOf(ratio);
				if (ratioCol == -1)
				{
					System.err.println("MergeBioAssay.extractAssays(BASEFileSection section, BASEFileReader reader)");
					System.err.println("Ratio error: Cant find annotation type " + ratio);
					System.exit(0);
				}
			}
		}

		try
		{
			String[] vals = reader.readDataRow(columns.size());
			while (vals != null)
			{
				Merge_assay tmp = new Merge_assay(vals[idCol], vals[nameCol]);
				if (annSort)
					tmp.v[sortnbr] = vals[sortCol];
				if (param_mergeMethod == RATIO_RATIO)
				{
					if (ratioCol != -1)
						tmp.v[11] = vals[ratioCol];
					else if (nbr != -1)
					{
						tmp.parse_name(param_assayname);
						tmp.v[11] = tmp.v[nbr];
					}
				}

				// sort the data into the right group.
				// if the assay lacks sort value (ie. sortvalue == "")
				// the data is placed in noGroup
				//  && (annType.isEmpty() || annType.contains(tmp.v[sortnbr]))
				if (annSort || tmp.parse_name(param_assayname))
				{
					if (tmp.v[sortnbr] != null)
					{
						List<Merge_assay> vec = assayGroups.get(tmp.v[sortnbr]);
						if (vec == null)
						{
							vec = new ArrayList<Merge_assay>();
							vec.add(tmp);
							assayGroups.put(tmp.v[sortnbr], vec);
						}
						else
							vec.add(tmp);
					}
					else
					{
						noGroup.add(tmp);
					}
				}
				else
				{
					noGroup.add(tmp);
				}
				vals = reader.readDataRow(columns.size());
			}
		}
		catch (BadSectionException e)
		{
			e.printStackTrace();
			return false;
		}
		catch (BadPatternException e)
		{
			e.printStackTrace();
			System.exit(0);
		}
		catch (IOException e)
		{
			e.printStackTrace();
			System.exit(0);
		}

		if (noGroup.size() > 0)
		{
			System.err.println(noGroup.size() + " didn't match the assayname " + param_assayname);
		}

		if (param_mergeMethod == RATIO_RATIO)
		{
			Iterator it = assayGroups.keySet().iterator();
			Object key;
			while (it.hasNext())
//			for (String key : assayGroups.keySet())
			{
				key = it.next();
				List<Merge_assay> group = assayGroups.get(key);
				if (group.size() != 2)
				{
					assayGroups.remove(key);
					noGroup.addAll(group);
					it = assayGroups.keySet().iterator();
				}
				else
				{
					Merge_assay assay1 = group.get(0);
					Merge_assay assay2 = group.get(1);

					if (!(assay1.v[11].equals(ratioNumerator) || assay1.v[11].equals(ratioDenominator)))
					{
						assayGroups.remove(key);
						noGroup.add(assay1);
						it = assayGroups.keySet().iterator();
					}
					if (!(assay2.v[11].equals(ratioNumerator) || assay2.v[11].equals(ratioDenominator)))
					{
						assayGroups.remove(key);
						noGroup.add(assay2);
						it = assayGroups.keySet().iterator();
					}
				}
			}
		}

		nbrOfGroups = assayGroups.size();
		if (nbrOfGroups == 0)
		{
			System.err.print("No groups found");
			System.exit(0);
		}
		return true;
	}

	/**
	 * Will read the spots from the baseFile
	 * 
	 * @param section - the section to be read. Must have type "spots"
	 * @param reader - the reader where the section data is located
	 * @return true if the baseFile is read without any problems
	 */
	public boolean mergeSpots(BASEFileSection section, BASEFileReader reader)
	{

		System.out.print("BASEfile\n");
		System.out.print("section\tassays\n" + "count\t" + (nbrOfGroups + noGroup.size()) + "\n");
		System.out.print("columns\tid\tname\tparents\n");
		System.out.print("%\n");
		int assay_nbr = 1;
		for (String key : assayGroups.keySet())
		{
			System.out.print(assay_nbr++ + "\t" + param_newname);
			List<Merge_assay> group = assayGroups.get(key);
			for (int i = 0; i < group.size(); i++)
			{
				Merge_assay tmp = group.get(i);
				if (i != 0)
				{
					System.out.print("/");
				}
				else
				{
					System.out.print(tmp.v[sortnbr] + "\t");
				}
				System.out.print(tmp.id);
			}
			System.out.print("\n");
		}

		for (int i = 0; i < noGroup.size(); i++)
		{
			BioAssay tmp = noGroup.get(i);
			System.out.print(tmp.id + "\t" + tmp.name + "\t" + tmp.id + "\n");
		}
		System.out.println();

		Merger m = null;
		switch (param_mergeMethod)
		{
			case GEOMETRIC_RATIO:
				m = new GeometricRatioMerger();
				break;
			case ARITHMETIC_RATIO:
				m = new ArithmeticRatioMerger();
				break;
			case ARITHMETIC_INT:
				m = new ArithmeticIntensityMerger();
				break;
			case RATIO_RATIO:
				m = new RatioRatioMerger(ratioNumerator, ratioDenominator);
				break;
		}
		m.merge(section, reader, assayGroups);
		return true;
	}

	/**
	 * A private utility method to split a string around a char
	 * 
	 * @param str - the string to split
	 * @param token - the token to use to split
	 * @return the strings as an <code>String[]</code>
	 */
	private String[] split(String str, char token)
	{
		ArrayList<String> ret = new ArrayList<String>();
		int start = 0, end = 0;

		while (str.length() > start)
		{
			end = str.indexOf(token, start);
			ret.add(str.substring(start, end));
			start = end + 1;
		}
		return ret.toArray(new String[ret.size()]);
	}
}