/* (non-javadoc)
 * GeometricRatioMerger.java is part of Merge.
 *
 * Merge is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Merge is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Merge; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * johan.enell@onk.lu.se
 * Johan Enell, Dept Oncology, Lund University, S-221 85 Lund, Sweden
 */
package mergebioassay.mergers;

import mergebioassay.BioAssay;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import basefile.BASEFileReader;
import basefile.BASEFileSection;
import basefile.BadSectionException;

/**
 * @author Johan Enell
 * 
 * The GeometricRatioMerger implements the abstract merger class. It will use
 * the geometric mean of the ratios.
 */
public class GeometricRatioMerger extends Merger
{
	@Override
	public void merge(BASEFileSection section, BASEFileReader reader, HashMap<String, List<Merge_assay>> assayGroups)
	{
		int intensity1Pos = -1;
		int intensity2Pos = -1;

		super.init(section, reader);

		intensity1Pos = assayFieldsMap.indexOf("intensity1");
		if (intensity1Pos == -1)
		{
			System.err.print("intensity1 field in assayFields row missing\n");
			System.exit(0);
		}

		intensity2Pos = assayFieldsMap.indexOf("intensity2");
		if (intensity2Pos == -1)
		{
			System.err.print("intensity2 field in assayFields row missing\n");
			System.exit(0);
		}

		try
		{
			System.out.print("section\tspots\n");
			System.out.print("columns\tposition\treporter\tassayData\n");
			System.out.print("assayFields\tl2ratio1_2\tl10intgmean1_2\tJE_nbrOfElements\n");
			System.out.print("setExtraFloats\tJE_nbrOfElements\n");
			System.out.print("assays");
			for (int i = 0; i < assayGroups.size(); i++)
				System.out.print("\t" + (i + 1));
			System.out.print("\n%\n");

			String[] row = reader.readDataRow(nbrOfColumns);

			while (row != null)
			{
				System.out.print(row[positionPos] + "\t" + row[reporterPos]);

				Iterator it = assayGroups.keySet().iterator();

				while (it.hasNext())
				{
					String key = (String) it.next();
					List micro = assayGroups.get(key);
					double int1 = 0, int2 = 0;
					DoubleVector ratioVector = new DoubleVector();
					DoubleVector aVector = new DoubleVector();
					for (int i = 0; i < micro.size(); i++)
					{
						BioAssay tmp = (BioAssay) (micro.get(i));
						int assay = assaysMap.indexOf(tmp.id);
						try
						{
							int1 = Double.parseDouble(row[assayDataPos + intensity1Pos + assay * assayFieldsMap.size()]);
							int2 = Double.parseDouble(row[assayDataPos + intensity2Pos + assay * assayFieldsMap.size()]);

							ratioVector.add(int1 / int2);
							aVector.add(Math.sqrt(int1 * int2));
						}
						catch (NumberFormatException nfe)
						{
							missingValues++;
						}
					}
					if (ratioVector.size() > 0)
					{

						System.out.print("\t" + log(geometricMean(ratioVector.toArray()), 2));
						System.out.print("\t" + log(geometricMean(aVector.toArray()), 10));
					}
					else
					{
						System.out.print("\t\t");
					}
					if (ratioVector.size() > 0) System.out.print("\t" + ratioVector.size());
					else System.out.print("\t");
				}
				System.out.println();
				row = reader.readDataRow(nbrOfColumns);
			}
		}
		catch (BadSectionException e)
		{
			e.printStackTrace();
		}
		catch (IOException e)
		{
			e.printStackTrace();
			System.exit(0);
		}
	}

	private float log(double d, int base)
	{
		return (float) (Math.log(d) / Math.log(base));
	}
}