/*
 * (non-javadoc) Merger.java is part of Merge.
 * 
 * Merge is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Merge is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Merge; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA
 * 
 * johan.enell@onk.lu.se Johan Enell, Dept Oncology, Lund University, S-221 85
 * Lund, Sweden
 */
package mergebioassay.mergers;

import java.util.HashMap;
import java.util.List;

import basefile.BASEFileReader;
import basefile.BASEFileSection;


/**
 * @author Johan Enell
 * 
 * Abstract class for merging BioAssays. The only method that the sybclass must
 * implement is the merge metod.
 */
public abstract class Merger
{
	protected int missingValues = 0;

	protected int nbrOfColumns = -1;

	protected int positionPos = -1;

	protected int assayDataPos = -1;

	protected int reporterPos = -1;

	protected List<String> assayFieldsMap;

	protected List<String> columnsMap;

	protected List<String> assaysMap;

	/**
	 * Will merge the bioassays in the groups and print the result on <code>System.out</code>.
	 * 
	 * @param section - the spot section with the parameters
	 * @param reader - a baseFileReader to read the data
	 * @param assayGroups - the group to be merged
	 */
	public void merge(BASEFileSection section, BASEFileReader reader, HashMap<String, List<Merge_assay>> assayGroups)
	{}

	/**
	 * Will initiate the merger by reading the parameters used in the merge
	 * method. Usally called in the start of the merge method.
	 * 
	 * @param section - the spot section with the parameters
	 * @param reader - a baseFileReader to read the data
	 */
	public void init(BASEFileSection section, BASEFileReader reader)
	{
		assaysMap = section.findFieldList("assays");
		if (assaysMap == null)
		{
			System.err.print("assaysMap row in spots section not found\n");
			System.exit(0);
		}

		columnsMap = section.findFieldList("columns");
		if (columnsMap == null)
		{
			System.err.print("Columns row in spots section not found\n");
			System.exit(0);
		}

		assayFieldsMap = section.findFieldList("assayFields");
		if (assayFieldsMap == null)
		{
			System.err.print("assayFields row in spots section not found\n");
			System.exit(0);
		}
		nbrOfColumns = (columnsMap.size() - 1) + (assayFieldsMap.size() * assaysMap.size());

		positionPos = columnsMap.indexOf("position");
		if (positionPos == -1)
		{
			System.err.print("position column in columns row missing\n");
			System.exit(0);
		}

		reporterPos = columnsMap.indexOf("reporter");
		if (reporterPos == -1)
		{
			System.err.print("reporter column in columns row missing\n");
			System.exit(0);
		}

		assayDataPos = columnsMap.indexOf("assayData");
		if (assayDataPos == -1)
		{
			System.err.print("assayData column in columns row missing\n");
			System.exit(0);
		}
	}

	/**
	 * Used to calculate the geometric mean.
	 * 
	 * @param values - the values to calculate mean
	 * @return the geometrc mean
	 */
	protected static double geometricMean(double[] values)
	{
		double ret = values[0];

		for (int i = 1; i < values.length; i++)
			ret *= values[i];

		ret = (Math.pow(ret, (1.0 /  values.length)));
		return ret;
	}

	/**
	 * Used to calculate the arithmetic mean.
	 * 
	 * @param values - the values to calculate mean
	 * @return the geometrc mean
	 */
	protected static double arithmeticMean(double[] values)
	{
		double ret = values[0];

		for (int i = 1; i < values.length; i++)
			ret += values[i];

		ret /= values.length;
		return ret;
	}

	/**
	 * Return the missing values.
	 * 
	 * @return the missing values
	 */
	public int getMissingValues()
	{
		return missingValues;
	}

	/**
	 * An internal class used to hold double values. Works basically as a Vector
	 * but is specialiced to hold the primitiv type <code>double</code>.
	 * 
	 * @author Johan Enell
	 */
	protected class DoubleVector
	{
		private double[] elementData;

		private int elementCount;

		private int capacityIncrement;

		/**
		 * Constructs a new DoubleVector.
		 */
		public DoubleVector()
		{
			elementData = new double[10];
			elementCount = 0;
			capacityIncrement = 2;
		}

		/**
		 * Appends the specified element to the end of this vector.
		 * 
		 * @param element - element to be added to this Vector
		 */
		public void add(double element)
		{
			if (elementCount < elementData.length)
			{
				elementData[elementCount] = element;
				elementCount++;
			}
			else
			{
				increaseCapacity();
				add(element);
			}
		}

		/**
		 * Returns an double array containing all of the elements in this list in
		 * proper sequence.
		 * 
		 * @return an array containing all of the elements in this list in proper
		 *         sequence.
		 */
		public double[] toArray()
		{
			double[] ret = new double[elementCount];
			for (int i = 0; i < elementCount; i++)
				ret[i] = elementData[i];
			return ret;
		}

		/**
		 * Returns the number of elements in this list.
		 * 
		 * @return the number of elements of this list.
		 */
		public int size()
		{
			return elementCount;
		}

		/**
		 * Will increase the capacity.
		 *  
		 */
		private void increaseCapacity()
		{
			double[] tmp = new double[elementData.length * capacityIncrement];
			for (int i = 0; i < elementCount; i++)
				tmp[i] = elementData[i];
			elementData = tmp;
		}
	}
}
